package psplot;

import java.awt.*;
import real.*;
import psplot.*;

// handles all coordinate transformations for a plotting canvas

class PlotCTM {
	int width, height;
	double m[][];
	double t[];
	final static double INFTY = 32783.0;
	
	// default coord system: in pixels, lower left = origin

	public PlotCTM(int w, int h) {
		width = w; height = h;
		m = new double[2][2];
		t = new double[2];
		m[0][0] = 1.0; m[0][1] =  0.0;
		m[1][0] = 0.0; m[1][1] = -1.0; 
		t[0] = 0.0;
		t[1] = (double) h;
	}

	public PlotCTM(PlotCTM ctm) {
		width = ctm.width;
		height = ctm.height;
		m = new double[2][2];
		for (int i=0;i<2;i++) {
			for (int j=0;j<2;j++) m[i][j] = ctm.m[i][j];
		}
		t = new double[2]; 
		t[0] = ctm.t[0];
		t[1] = ctm.t[1];
	}

	void setCorners(int w, int h, 
			double llx, double lly, double urx, double ury) {
		width = w; height = h;
		setCorners(llx, lly, urx, ury);
	}

	void setCorners(double llx, double lly, double urx, double ury) {
		m[0][0] = (double) width/(urx - llx);
		m[0][1] = 0.0;
		m[1][1] = - (double) height/(ury - lly);
		m[1][0] = 0.0;
		t[0] = - m[0][0]*llx;
		t[1] = - m[1][1]*ury;
	}

	// coords in can be infty but not NaN
	// all points returned have real numbers - +infty etc.

	// coords to pixels
	Point toPoint(RealPoint p) {
		return(toPoint(p.x, p.y));
	}
	
	Point toPoint(double x, double y) {
		double x0 = m[0][0]*x + m[0][1]*y + t[0];
		double y0 = m[1][0]*x + m[1][1]*y + t[1];
		int x1, y1;
		// never let it wrap around
		if (-INFTY < x0 && x0 < INFTY) {
			x1 = (int) Math.round(x0);
		}
		else if (x0 <= -INFTY || x0 == Double.NEGATIVE_INFINITY) {
			x1 = (int) -INFTY;
		}
		else {
			x1 = (int) INFTY;
		}
		if (-INFTY < y0 && y0 < INFTY) {
			y1 = (int) Math.round(y0);
		}
		else if (y0 <= -INFTY || y0 == Double.NEGATIVE_INFINITY) {
			y1 = (int) -INFTY;
		}
		else {
			y1 = (int) INFTY;
		}
		Point q = new Point(x1, y1);
		return(q);
	}

	// pixels to coords
	public RealPoint toRealPoint(Point p) {
		return(toRealPoint(p.x, p.y));
	}

	// pixels to coords
	// p = m*q + t, q = m^{-1}(p - t)
	public RealPoint toRealPoint(int x, int y) {
		double x0 = (double) x - t[0];
		double y0 = (double) y - t[1];
		double d = m[0][0]*m[1][1] - m[0][1]*m[1][0];
		double x1 =  m[1][1]*x0 - m[0][1]*y0;
		double y1 = -m[1][0]*x0 + m[0][0]*y0;
		x1 /= d;
		y1 /= d;
		RealPoint q = new RealPoint(x1, y1);
		return(q);
	}

	// coords to pixels but in real numbers
	RealPoint toFinePoint(RealPoint p) {
		return(toFinePoint(p.x, p.y));
	}
	
	RealPoint toFinePoint(double x, double y) {
		double x0 = m[0][0]*x + m[0][1]*y + t[0];
		double y0 = m[1][0]*x + m[1][1]*y + t[1];
		RealPoint q = new RealPoint(x0, y0);
		return(q);
	}

	// returns false if x or y is NaN 
	// since all comparisons with NaN are false
	boolean isVisible(double x, double y) {
		Point p = toPoint(x, y);
		if (0 <= p.x && p.x < width 
			&& 0 <= p.y && p.y < height) {
			return(true);
		}
		else {
			return(false);
		}
	}

	boolean isVisible(RealPoint p) {
		return(isVisible(p.x, p.y));
	}

	boolean isVisible(Point p) {
		// doesn't use xmin etc.
		if (0 <= p.x && p.x < width 
			&& 0 <= p.y && p.y < height) {
			return(true);
		}
		else {
			return(false);
		}
	}

	// following the PostScript model
	// but using column vectors
	// all of these postmultiply the CTM by the relevant matrix 

	void scale(double a, double b) {
		m[0][0] *= a;
		m[1][0] *= a;
		m[0][1] *= b;
		m[1][1] *= b;
	}

	void translate(double a, double b) {
		t[0] += m[0][0]*a + m[0][1]*b;
		t[1] += m[1][0]*a + m[1][1]*b;
	}
	
	// multiply on the right by rotation matrix
	void rotate(double theta) {
		double c = Math.cos(theta);
		double s = Math.sin(theta);
		
		double x00 =  m[0][0]*c + m[0][1]*s;
		double x01 = -m[0][0]*s + m[0][1]*c;
		double x10 =  m[1][0]*c + m[1][1]*s;
		double x11 = -m[1][0]*s + m[1][1]*c;
		m[0][0] = x00;
		m[0][1] = x01;
		m[1][0] = x10;
		m[1][1] = x11;
	}

	// z = magnification, t = new origin
	void zoom(double z, RealPoint t) {
		Point s;
		RealPoint r;

		// calculate new matrix
		scale(z, z);
		s = new Point(width/2, height/2);
		r = toRealPoint(s);
		translate(r.x - t.x, r.y - t.y);
	}

	void zoom(double mx, double my, RealPoint t) {
		Point s;
		RealPoint r;

		// calculate new matrix
		scale(mx, my);
		s = new Point(width/2, height/2);
		r = toRealPoint(s);
		translate(r.x - t.x, r.y - t.y);
	}

	public String toString() {
		return("ctm: " + Real.toString(m[0][0]) + " " 
				+ Real.toString(m[0][1]) + '\n'
				+ Real.toString(m[1][0]) + " " 
				+ Real.toString(m[1][1]) + '\n' 
				+ Real.toString(t[0]) + " " 
				+ Real.toString(t[1]) + '\n');
	}
}

